from django import forms
from django.contrib.auth import get_user_model

User = get_user_model()


class LoginForm(forms.Form):
    username = forms.CharField()
    password = forms.CharField(widget=forms.PasswordInput)

    def clean(self):
        super(LoginForm, self).clean()
        username = self.cleaned_data.get('username').lower()
        password = self.cleaned_data.get('password')
        user = User.objects.filter(username=username).first()

        if not user:
            raise forms.ValidationError('User is not found')

        if user and not user.is_active:
            raise forms.ValidationError('User is inactive. You must confirm the registration email address at registration.')

        valid = user.check_password(password)
        if not valid:
            raise forms.ValidationError('Invalid: username / password')
        return valid



from rest_framework.authentication import TokenAuthentication
from django.contrib.auth import get_user_model
from django.utils import timezone
from django.conf import settings
from datetime import timedelta
from garpix_user.utils.get_token_from_request import get_token_from_request
from django.contrib.auth.models import AnonymousUser


def get_user_by_token(token):
    from ..models.access_token import AccessToken as Token
    from oauth2_provider.models import AccessToken
    User = get_user_model()

    # Refresh django rest token
    try:
        tok = Token.objects.get(key=token)
        if settings.GARPIX_ACCESS_TOKEN_TTL_SECONDS > 0:
            if tok.created + timedelta(seconds=settings.GARPIX_ACCESS_TOKEN_TTL_SECONDS) < timezone.now():
                tok.delete()
                raise Exception("Token expired.")
        user = User.active_objects.get(id=tok.user_id)
        return user
    except: # noqa
        pass

    # Refresh social auth token
    try:
        tok = AccessToken.objects.get(token=token)
        if settings.GARPIX_ACCESS_TOKEN_TTL_SECONDS > 0:
            if tok.created + timedelta(seconds=settings.GARPIX_ACCESS_TOKEN_TTL_SECONDS) < timezone.now():
                tok.delete()
        else:
            user = tok.user
            return user
    except: # noqa
        pass

    return AnonymousUser()


class MainAuthentication(TokenAuthentication):
    keyword = 'Bearer'

    def authenticate(self, request):
        token = get_token_from_request(request, keyword=self.keyword)
        if token is None:
            return None

        user = get_user_by_token(token)
        if user is not None:
            return user, None
        return user, None



from django.contrib import admin
from garpix_user.models import UserSession


@admin.register(UserSession)
class UserSessionAdmin(admin.ModelAdmin):
    fields = ('user', 'token_number', 'recognized', 'last_access', 'is_phone_confirmed', 'is_email_confirmed')



from rest_framework import serializers


class ToLowerMixin(serializers.Serializer):

    def validate_email(self, value):
        """
        ensure that email is always lower case.
        """
        return value.lower()

    def validate_username(self, value):
        """
        ensure that username is always lower case.
        """
        return value.lower()



from rest_framework import serializers

from garpix_user.mixins.serializers import ToLowerMixin, PasswordSerializerMixin
from django.utils.translation import gettext_lazy as _


class RestorePasswordSerializer(ToLowerMixin, serializers.Serializer):
    username = serializers.CharField(required=True)


class RestoreSetPasswordSerializer(PasswordSerializerMixin, serializers.Serializer):
    new_password = serializers.CharField(max_length=255, required=True)
    username = serializers.CharField(required=True)
    restore_password_confirm_code = serializers.CharField(max_length=15, required=True)

    def validate_new_password(self, value):
        self._validate_password(value)
        return value


class RestoreCheckCodeSerializer(serializers.Serializer):
    restore_password_confirm_code = serializers.CharField(max_length=15, required=True)


class ChangePasswordSerializer(PasswordSerializerMixin, serializers.Serializer):
    new_password = serializers.CharField(max_length=255, required=True)
    password = serializers.CharField(required=True)

    def validate_new_password(self, value):
        self._validate_password(value)
        return value

    def validate_password(self, value):
        user = self.context['request'].user
        if not user.check_password(value):
            raise serializers.ValidationError(
                _('Password is incorrect')
            )
        return value



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from garpix_utils.string import get_random_string
from rest_framework import serializers

from garpix_user.mixins.serializers import PasswordSerializerMixin
from garpix_user.models.user_session import UserSession
from django.utils.translation import ugettext as _

User = get_user_model()


class RegistrationSerializer(PasswordSerializerMixin, serializers.ModelSerializer):
    password = serializers.CharField(write_only=True)
    password_2 = serializers.CharField(write_only=True)

    def validate_password_2(self, value):

        # check for password == password_2
        if value != self.initial_data.get('password'):
            raise serializers.ValidationError(_("Passwords do not match"))

        return value

    def validate_password(self, value):

        self._validate_password(value)

        return value

    def validate_email(self, value):

        GARPIX_USER_SETTINGS = settings.GARPIX_USER

        request = self.context.get('request')

        queryset = User.objects.filter(email=value).first()
        if queryset is not None:
            raise serializers.ValidationError(_("This email is already in use"))

        if GARPIX_USER_SETTINGS.get('USE_PREREGISTRATION_EMAIL_CONFIRMATION', False) and GARPIX_USER_SETTINGS.get(
                'USE_EMAIL_CONFIRMATION', False):
            user = UserSession.get_or_create_user_session(request)
            if not user.is_email_confirmed:
                raise serializers.ValidationError(_('Email was not confirmed'))

        return str(value).lower() if value else None

    def validate_phone(self, value):

        GARPIX_USER_SETTINGS = settings.GARPIX_USER

        request = self.context.get('request')

        queryset = User.objects.filter(phone=value).first()
        if queryset is not None:
            raise serializers.ValidationError(_("This phone is already in use"))

        if GARPIX_USER_SETTINGS.get('USE_PREREGISTRATION_PHONE_CONFIRMATION', False) and GARPIX_USER_SETTINGS.get(
                'USE_PHONE_CONFIRMATION', False):
            user = UserSession.get_or_create_user_session(request)
            if not user.is_phone_confirmed:
                raise serializers.ValidationError(_('Phone number was not confirmed'))

        return value

    def create(self, validated_data):

        GARPIX_USER_SETTINGS = settings.GARPIX_USER

        request = self.context.get('request', None)

        validated_data.pop('password_2')

        user_data = validated_data

        if 'username' not in User.USERNAME_FIELDS and 'username' not in validated_data.keys():
            user_data.update({'username': get_random_string(25)})

        if GARPIX_USER_SETTINGS.get('USE_PHONE_CONFIRMATION', False) and not GARPIX_USER_SETTINGS.get('USE_PREREGISTRATION_PHONE_CONFIRMATION', False):
            user_data.update({'is_phone_confirmed': False})

        if GARPIX_USER_SETTINGS.get('USE_EMAIL_CONFIRMATION', False) and not GARPIX_USER_SETTINGS.get('USE_PREREGISTRATION_EMAIL_CONFIRMATION', False):
            user_data.update({'is_email_confirmed': False})

        with transaction.atomic():
            user = User.objects.create_user(**user_data)
            if request:
                user_session = UserSession.get_or_create_user_session(request)
                user_session.user = user
                user_session.recognized = UserSession.UserState.REGISTERED
                user_session.save()

        return user

    class Meta:
        model = User
        fields = ('password', 'password_2',)

    def get_field_names(self, declared_fields, info):
        expanded_fields = super().get_field_names(declared_fields, info)

        if getattr(User, 'USERNAME_FIELDS', None):
            return expanded_fields + User.USERNAME_FIELDS
        else:
            return expanded_fields

    def get_extra_kwargs(self):
        extra_kwargs = super().get_extra_kwargs()
        if USERNAME_FIELDS := getattr(User, 'USERNAME_FIELDS', None):
            for field in USERNAME_FIELDS:
                extra_kwargs.update({field: {'required': True, 'allow_blank': False}})

        return extra_kwargs



from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers
from rest_framework.authentication import authenticate

from garpix_user.models import UserSession


class AuthTokenSerializer(serializers.Serializer):
    username = serializers.CharField(label=_("Username"))
    password = serializers.CharField(
        label=_("Password"),
        style={'input_type': 'password'},
        trim_whitespace=False
    )

    def validate(self, attrs):
        username = attrs.get('username')
        password = attrs.get('password')
        username = username.lower()

        if username and password:
            request = self.context.get('request')
            user = authenticate(request=request,
                                username=username, password=password)
            # The authenticate call simply returns None for is_active=False
            # users. (Assuming the default ModelBackend authentication
            # backend.)
            if not user:
                msg = _('Unable to log in with provided credentials.')
                raise serializers.ValidationError(msg, code='authorization')
            user.set_user_session(request)
        else:
            msg = _('Must include "username" and "password".')
            raise serializers.ValidationError(msg, code='authorization')

        attrs['user'] = user
        return attrs



import uuid
from django.db import models
from django.conf import settings
from django.db.models import Q
from django.utils import timezone
from django.contrib.auth import get_user_model
from django.utils.translation import ugettext as _
from phonenumber_field.modelfields import PhoneNumberField

from garpix_user.mixins.models import RestorePasswordMixin
from garpix_user.mixins.models.confirm import UserEmailConfirmMixin, UserPhoneConfirmMixin


class UserSession(RestorePasswordMixin, UserEmailConfirmMixin, UserPhoneConfirmMixin, models.Model):
    HEAD_NAME = 'user-session-token'

    class UserState(models.IntegerChoices):
        UNRECOGNIZED = (0, _('Undefined'))
        GUEST = (1, _('Guest'))
        REGISTERED = (2, _('Registered'))

    email = models.EmailField(verbose_name='Email', null=True, blank=True)
    is_email_confirmed = models.BooleanField(_("Email confirmed"), default=False)

    phone = PhoneNumberField(_("Phone number"), blank=True, default='')
    is_phone_confirmed = models.BooleanField(_("Phone number confirmed"), default=False)

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        verbose_name=_('User')
    )
    token_number = models.CharField(max_length=256, null=True, blank=True, verbose_name='user token')
    recognized = models.PositiveIntegerField(
        default=UserState.UNRECOGNIZED,
        choices=UserState.choices,
        verbose_name=_('Type'),
        help_text=_('Indicates the state in which the user is recognized.')
    )
    last_access = models.DateTimeField(
        verbose_name=_('Last entrance'),
        default=timezone.now,
    )

    @classmethod
    def get_from_request(cls, request):
        user = request.user
        if user.is_authenticated:
            return UserSession.objects.filter(user=user).first()

        token = request.headers.get(cls.HEAD_NAME, None)

        if token is not None:
            user_session, _res = UserSession.objects.get_or_create(token_number=token)
            return user_session

        token = request.session.session_key
        if token is not None:
            return UserSession.objects.filter(token_number=token).first()

        username = request.GET.get('username', None)
        if username is not None:
            query = Q()
            for field in get_user_model().USERNAME_FIELDS:
                query |= Q(**{f'user__{field}': username.lower()})
            return UserSession.objects.filter(query).first()
        return None

    @classmethod
    def create_from_request(cls, request, username, session):

        User = get_user_model()

        if request.user.is_authenticated:
            user = User.objects.get(pk=request.user.pk)
            return UserSession.objects.create(
                user=user,
                token_number=uuid.uuid4(),
                recognized=UserSession.UserState.REGISTERED
            )

        if session is True:
            token = request.session.session_key
            return UserSession.objects.create(
                token_number=token,
                recognized=UserSession.UserState.GUEST
            )

        if username is not None:
            query = Q()
            for field in User.USERNAME_FIELDS:
                query |= Q(**{field: username.lower()})

            try:
                user = User.objects.get(query)
                session_user = UserSession.objects.filter(user=user).first()
                return session_user or UserSession.objects.create(
                    token_number=uuid.uuid4(),
                    recognized=UserSession.UserState.REGISTERED,
                    user=user
                )
            except Exception as e:
                print(e)

        return UserSession.objects.create(
            token_number=uuid.uuid4(),
            recognized=UserSession.UserState.GUEST
        )

    @classmethod
    def set_user_from_request(cls, request):
        user_session = cls.get_from_request(request)
        if request.user.is_authenticated and user_session is not None:
            user = get_user_model().objects.get(pk=request.user.pk)
            user_session.user = user
            user_session.save()
            return True
        return False

    @classmethod
    def get_or_create_user_session(cls, request, username=None, session=False):

        user_session = cls.get_from_request(request)
        if user_session is not None:
            return user_session

        return cls.create_from_request(request, username, session)

    class Meta:
        verbose_name = _('System user')
        verbose_name_plural = _('System users')

    def __str__(self):
        return _(
            f"{self.UserState.choices[self.recognized][1]} {self.user.username}" if self.user else f'{self.UserState.choices[self.recognized][1]} № {self.pk}')



def get_token_from_request(request, keyword='Bearer'):
    if 'HTTP_AUTHORIZATION' not in request.META:
        return None

    token = request.META['HTTP_AUTHORIZATION']
    token = token[len(keyword) + 1:]
    return token



from django.conf import settings
from django.contrib.auth import get_user_model
from django.views.generic import RedirectView

from garpix_user.serializers import EmailConfirmSendSerializer, EmailConfirmCheckCodeSerializer, \
    EmailPreConfirmSendSerializer, UserSessionTokenSerializer

from drf_spectacular.utils import extend_schema
from rest_framework import viewsets, status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.exceptions import NotAuthenticated, ValidationError
from garpix_user.models import UserSession
from django.utils.translation import ugettext as _

from garpix_user.utils.drf_spectacular import user_session_token_header_parameter


@extend_schema(
    parameters=[
        user_session_token_header_parameter()
    ]
)
class EmailConfirmationView(viewsets.GenericViewSet):

    def get_serializer_class(self):
        user = self.request.user
        if self.action == 'send_code':
            if user.is_authenticated:
                return EmailConfirmSendSerializer
            return EmailPreConfirmSendSerializer
        return EmailConfirmCheckCodeSerializer

    @extend_schema(summary=_('Email confirmation. Step 1' if not settings.GARPIX_USER.get('USE_EMAIL_LINK_CONFIRMATION',
                                                                                          False) else _('Send email confirmation link')))
    @action(methods=['POST'], detail=False)
    def send_code(self, request, *args, **kwargs):
        user = request.user
        if user.is_authenticated:
            serializer = self.get_serializer(data=request.data)
            serializer.is_valid(raise_exception=True)

            result = user.send_email_confirmation_code(serializer.data.get('email', None))
            if result is not True:
                result.raise_exception(exception_class=ValidationError)
            return Response({'result': 'success'})
        else:
            if settings.GARPIX_USER.get('USE_PREREGISTRATION_EMAIL_CONFIRMATION', False):
                user = UserSession.get_or_create_user_session(request)
                serializer = self.get_serializer(data=request.data)
                serializer.is_valid(raise_exception=True)

                result = user.send_email_confirmation_code(serializer.data['email'])
                if result is not True:
                    result.raise_exception(exception_class=ValidationError)
                return Response({'result': 'success'})

            raise NotAuthenticated()

    if not settings.GARPIX_USER.get('USE_EMAIL_LINK_CONFIRMATION', True):
        @extend_schema(summary=_('Email confirmation. Step 2'))
        @action(methods=['POST'], detail=False)
        def check_code(self, request, *args, **kwargs):
            user = request.user
            if not user.is_authenticated:
                if settings.GARPIX_USER.get('USE_PREREGISTRATION_EMAIL_CONFIRMATION', False):
                    user = UserSession.get_or_create_user_session(request)
                else:
                    raise NotAuthenticated()

            serializer = self.get_serializer(data=request.data)
            serializer.is_valid(raise_exception=True)
            result = user.confirm_email(serializer.data['email_confirmation_code'])
            if result is not True:
                result.raise_exception(exception_class=ValidationError)
            return Response({'result': 'success'})


class EmailConfirmationLinkView(RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        User = get_user_model()
        hash = self.kwargs.get('hash', None)
        result = User.confirm_email_by_link(hash)
        if result is not True:
            return f"{settings.GARPIX_USER.get('EMAIL_CONFIRMATION_LINK_REDIRECT', '/')}?status=error"
        return f"{settings.GARPIX_USER.get('EMAIL_CONFIRMATION_LINK_REDIRECT', '/')}?status=success"



from typing import Optional

from drf_spectacular.utils import extend_schema
from rest_framework import parsers
from rest_framework import permissions
from rest_framework import status
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from garpix_user.models import UserSession
from ..serializers import UserSessionTokenSerializer
from ..utils.drf_spectacular import user_session_token_header_parameter


@extend_schema(
    parameters=[
        user_session_token_header_parameter()
    ]
)
class UserSessionView(viewsets.ViewSet):
    parser_classes = (parsers.JSONParser,)
    permission_classes = (permissions.AllowAny,)

    def get_user_session(self) -> Optional[UserSession]:
        return UserSession.get_from_request(self.request)

    def get_or_create_user_session(self, session=False):
        return UserSession.get_or_create_user_session(request=self.request, session=session)

    @action(detail=False, methods=['POST'], permission_classes=(permissions.AllowAny,))
    def create_user_session(self, request):
        return Response({
            'session_user': UserSessionTokenSerializer(self.get_or_create_user_session()).data
        }, status=status.HTTP_200_OK)
